from __future__ import annotations

from collections import Counter
from typing import AbstractSet, Mapping, MutableMapping

from pyrsistent import pmap, pset
from pyrsistent.typing import PMap as PMapType, PSet as PSetType

from carladam.petrinet.color import ColorSet
from carladam.petrinet.place import Place
from carladam.petrinet.token import Token

Marking = Mapping[Place, AbstractSet[Token]]
"A mapping between `Place`s and their respective `Token`s."

MutableMarking = MutableMapping[Place, AbstractSet[Token]]

PMarking = PMapType[Place, PSetType[Token]]


def marking_colorset(marking: Marking) -> Mapping[Place, ColorSet]:
    """
    Transforms a `Marking` into the quantities of tokens (grouped by color) for each `Place`.

    Empty places are not included in the return value.
    This ensures brevity in unit tests when comparing with expected values.
    """
    return pmap((place, pmap(Counter(token.color for token in tokens))) for place, tokens in marking.items() if tokens)


def pmarking(marking: Marking | MutableMarking | PMarking) -> PMarking:
    """Returns an immutable marking given a mutable or immutable marking."""
    return pmap((place, pset(tokens)) for place, tokens in marking.items())



from __future__ import annotations

from itertools import islice
from typing import FrozenSet, Iterator, Mapping, MutableMapping, Protocol, Sequence, TYPE_CHECKING, Tuple

from attr import define
from pyrsistent import pmap

from carladam.petrinet import defaults

if TYPE_CHECKING:  # pragma: nocover
    from carladam.petrinet.token import Token, TokenSet
    from carladam.petrinet.transition import TransitionFunction


@define
class Color:
    """A `Color` is a discrete label used to differentiate types of `Token`s."""

    label: str
    "Globally unique name of this `Color`."

    def __repr__(self):
        return self.label

    def __hash__(self):
        """Colors with the same name are the same color."""
        return hash(self.label)

    def __call__(self, **kwargs) -> Token:
        """Return a `Token` of this `Color`, having its `data` set to the passed-in `kwargs`."""
        from carladam.petrinet.token import Token

        return Token(color=self, data=kwargs)

    def passthrough(self, quantity: int = 1) -> TransitionFunction:
        """Return a transition function that passes through (unmodified) tokens of this color."""
        from carladam.petrinet.transition import passthrough

        return passthrough({self: quantity})

    def token_generator(self, data: Mapping) -> Iterator[Token]:
        """Return a generator that produces tokens of this color with the given `data`."""
        from carladam.petrinet.token import Token

        pdata = pmap(data)
        while True:
            yield Token(color=self, data=pdata)

    def produce(self, quantity: int = 1, **kwargs) -> TransitionFunction:
        """Return a transition function that generates a given quantity of this color of token."""

        def transition_fn(inputs: TokenSet) -> Iterator[TokenSet]:
            yield frozenset(islice(self.token_generator(kwargs), quantity))

        return transition_fn


class HasColor(Protocol):
    color: Color


Abstract = Color(defaults.ABSTRACT_TOKEN)
"Abstract tokens are indicated by a filled circle."

Quantity = int
ColorSet = Mapping[Color, Quantity]
"""Mapping between a set of `Color`s and an expected quantity of tokens of that color.

These are used to specify

These are also used when transforming a `Marking` (mapping of `Place` → `Set[Token]`)
into a mapping of `Place` → (`Color` → `Quantity`).
"""

MutableColorSet = MutableMapping[Color, Quantity]
"""Pyrsistent version of ColorSet."""

ColorSetItem = Tuple[Color, Quantity]
"A single item in a ColorSet."

FrozenColorSet = FrozenSet[ColorSetItem]
"A frozenset transformation of a ColorSet."


def colorset_string(colorset: ColorSet) -> str:
    """Return a string representation of a `ColorSet` suitable for decorating an `Arc` in a diagram."""
    if colorset == {Abstract: 1}:
        return ""
    sorted_colorset: Sequence[tuple[str, Quantity]] = sorted(
        (color.label, quantity) for color, quantity in colorset.items()
    )
    return "".join(name * quantity for name, quantity in sorted_colorset)


def color_eq(color: Color):
    """Return a function that compares an object's `color` property with the given `color`."""

    def comparator(obj: HasColor):
        return obj.color == color

    return comparator



from __future__ import annotations


class PetriNetException(Exception):
    """Base class for all exceptions related to `PetriNet` operations."""


class PetriNetArcIncomplete(PetriNetException):
    """The arc cannot be added because its endpoints are not set."""


class PetriNetTransitionFunctionOutputHasOverlappingColorsets(PetriNetException):
    """The transition cannot occur due to overlapping colorsets of the output tokensets."""


class TransitionNotEnabled(Exception):
    pass


class TransitionHasNoArcs(TransitionNotEnabled):
    pass


class ArcGuardReturnsFalse(TransitionNotEnabled):
    pass


class ArcGuardRaisesException(Exception):
    pass


class TransitionGuardReturnsFalse(TransitionNotEnabled):
    pass


class TransitionGuardRaisesException(Exception):
    pass



"""Provides commonly used classes and functions when using Jupyter notebooks."""

from __future__ import annotations

from functools import lru_cache

import httpx
from IPython.display import HTML
from jinja2 import Template

from carladam import (
    Abstract,
    Annotate,
    Color,
    Marking,
    PetriNet,
    Place,
    Token,
    TransformEach,
    Transition,
    __,
    autoname,
    passthrough,
)
from carladam.diagram.digraph import graphviz_digraph
from carladam.diagram.kroki import kroki_image_url, niolesk_edit_url
from carladam.petrinet.defaults import EMPTY_SET

__all__ = [
    "Abstract",
    "Annotate",
    "Color",
    "Marking",
    "PetriNet",
    "Place",
    "Token",
    "TransformEach",
    "Transition",
    "__",
    "autoname",
    "passthrough",
    "report",
]

from carladam.petrinet.marking import pmarking


@lru_cache
def get_image_svg_text(image_url: str) -> str:
    return httpx.get(image_url).text


def report(net: PetriNet, marking: Marking | None = None, inline: bool = True) -> HTML:
    """Returns a report showing useful information about a `PetriNet` given a `Marking`."""
    marking = pmarking(marking or net.empty_marking())
    # Render diagrams.
    graph_source = graphviz_digraph(net, pmarking(marking), legend=True, legend_width=50)
    graph_image_url = kroki_image_url(diagram_source=graph_source, diagram_type="graphviz")
    graph_image_svg_content = get_image_svg_text(graph_image_url)
    graph_edit_url = niolesk_edit_url(graph_image_url)
    enabled_transitions = set(net.enabled_transitions(marking))
    subnets = {transition: net.subnet(transition) for transition in enabled_transitions}
    subgraph_sources = {
        transition: graphviz_digraph(subnet, marking, legend=True) for transition, subnet in subnets.items()
    }
    subgraph_image_urls = {
        transition: kroki_image_url(diagram_source=subgraph_source, diagram_type="graphviz")
        for transition, subgraph_source in subgraph_sources.items()
    }
    subgraph_image_svg_content = {
        transition: get_image_svg_text(subgraph_image_url)
        for transition, subgraph_image_url in subgraph_image_urls.items()
    }
    subgraph_edit_urls = {
        transition: niolesk_edit_url(subgraph_image_url)
        for transition, subgraph_image_url in subgraph_image_urls.items()
    }
    html = REPORT_TEMPLATE.render(
        EMPTY_SET=EMPTY_SET,
        enabled_transitions=enabled_transitions,
        graph_edit_url=graph_edit_url,
        graph_image_svg_content=graph_image_svg_content,
        graph_image_url=graph_image_url,
        inline=inline,
        subgraph_edit_urls=subgraph_edit_urls,
        subgraph_image_svg_content=subgraph_image_svg_content,
        subgraph_image_urls=subgraph_image_urls,
    )
    return HTML(html)


REPORT_TEMPLATE = Template(
    # language=jinja2
    """
        <div style="display: grid">
            <div style="grid-row: 1 / 2; grid-column: 1 / 3">
                <p><em>Entire net:</em></p>
                <a href="{{ graph_edit_url }}">
                    {% if inline %}
                        {{ graph_image_svg_content | safe }}
                    {% else %}
                        <img src="{{ graph_image_url }}" alt="Petri net graph" />
                    {% endif %}
                </a>
            </div>
            <div style="grid-row: 1 / 2; grid-column: 3 / 4">
                <p><em>Enabled transitions:</em>{% if not enabled_transitions %}{{ EMPTY_SET }}{% endif %}</p>
                {% if enabled_transitions %}
                    {% for transition in enabled_transitions %}
                        <div style="display: inline-block">
                            <a href="{{ subgraph_edit_urls[transition] }}">
                                {% if inline %}
                                    {{ subgraph_image_svg_content[transition] | safe }}
                                {% else %}
                                    <img src="{{ subgraph_image_urls[transition] }}" alt="Petri net subgraph" />
                                {% endif %}
                            </a>
                        </div>
                    {% endfor %}
                {% endif %}
            </div>
        </div>
    """
)



from functools import lru_cache
from typing import Mapping

from pyrsistent import freeze
from pyrsistent.typing import PMap

from carladam import Color, PetriNet, Token
from carladam.petrinet.marking import MutableMarking, PMarking, pmarking


def decode_marking_from_json(net: PetriNet, colors: Mapping[str, Color], marking_json: Mapping) -> PMarking:
    if not isinstance(marking_json, PMap):
        marking_json = freeze(marking_json)
    if not isinstance(colors, PMap):
        colors = freeze(colors)
    return _decode_marking_from_json(net, colors, marking_json)


@lru_cache
def _decode_marking_from_json(net: PetriNet, colors: Mapping[str, Color], marking_json: PMap) -> PMarking:
    _initial_marking: MutableMarking = {}
    for place in net.places:
        if place.id not in marking_json:
            continue
        place_marking = _initial_marking[place] = set()
        for token_json in marking_json[place.id]:
            token = Token(
                id=token_json["id"],
                color=colors[token_json["color"]],
                data=token_json["data"],
            )
            place_marking.add(token)
    return pmarking(_initial_marking)



from __future__ import annotations

from collections import Counter, defaultdict
from collections.abc import Iterable
from typing import AbstractSet, Callable, Iterator, Sequence, TYPE_CHECKING, cast, overload

from attr import define, field
from pyrsistent import PMap, PSet, pmap, pset

from carladam.petrinet import defaults
from carladam.petrinet.color import Abstract, Color, ColorSet, MutableColorSet
from carladam.petrinet.defaults import default_id
from carladam.petrinet.token import TokenSet

if TYPE_CHECKING:  # pragma: nocover
    from carladam.petrinet.arc import ArcPT, ArcTP, CompletedArcTP
    from carladam.petrinet.place import Place

TransitionGuard = Callable[[TokenSet], bool]
"Type of function receiving a set of `Token`s from input `Place`s and returning `True` if guard(s) are met."

TransitionFunction = Callable[[TokenSet], Iterator[TokenSet]]


def always(value: bool) -> TransitionGuard:
    """Returns a transition guard function that always returns `value`."""

    def transition_guard(inputs: TokenSet) -> bool:
        return value

    return transition_guard


def convert_fn_list_to_generator(fn: Iterator[TransitionFunction] | TransitionFunction) -> TransitionFunction:
    """Given a list of functions, return a generator that yields the results of calling each function."""
    if not isinstance(fn, Iterable):
        return fn

    def _fn(inputs: TokenSet) -> Iterator[TokenSet]:
        for inner_fn in cast(Iterator[TransitionFunction], fn):
            yield from inner_fn(inputs)

    return _fn


def passthrough(*colorsets: Color | ColorSet | AbstractSet) -> TransitionFunction:
    """
    Return a transition function that passes through inputs as outputs, optionally matching a colorset.

    The colorset used for matching is the superset of the colorsets passed in.
    """

    def _fn_all(inputs: TokenSet) -> Iterator[TokenSet]:
        yield pset(inputs)

    if not colorsets:
        return _fn_all

    passthrough_colorset_dict: MutableColorSet = defaultdict(int)
    for colorset in colorsets:
        if colorset and isinstance(colorset, Color):
            colorset = {colorset: 1}
        if colorset and isinstance(colorset, (set, frozenset, PSet)):
            colorset = {color: 1 for color in colorset}
        if not isinstance(colorset, (dict, PMap)):
            continue
        color: Color
        quantity: int
        for color, quantity in colorset.items():
            passthrough_colorset_dict[color] += quantity
    passthrough_colorset = pmap(passthrough_colorset_dict)

    def _fn(inputs: TokenSet) -> Iterator[TokenSet]:
        inputs_colorset = pmap(Counter(token.color for token in inputs))
        if inputs_colorset == passthrough_colorset:
            # Direct match.
            yield pset(inputs)
            return
        # Pass through a subset if we can.
        # Note: Token selection order is undefined if len of input tokens of a certain color
        # exceeds the corresponding colorset quantity.
        remaining = dict(passthrough_colorset)
        subset = set()
        for token in inputs:
            if token.color not in remaining:
                continue
            subset.add(token)
            remaining[token.color] -= 1
            if not remaining[token.color]:
                del remaining[token.color]
        if remaining:
            return
        yield pset(subset)

    return _fn


@define
class Transition:
    """When a `Transition` occurs it consumes `Token`s from `Place`s and produces `Token`s to `Place`s."""

    id: str = field(factory=default_id, repr=False)
    "Globally unique ID, used as a hash key."

    name: str = field()
    "Descriptive name, usually but not necessarily unique within a net."

    guard: TransitionGuard = field(default=always(True))
    "Function accepting a set of `Token`s and returning True if guard conditions are met."

    annotation: str | None = None
    "Text annotation to show on a diagram."

    fn: TransitionFunction | Sequence[TransitionFunction] = field(
        default=Abstract.produce(),
        repr=False,
        converter=convert_fn_list_to_generator,
    )
    "Function accepting a set of `Token`s to consume and returning a sequence of `TokenSet` generators."

    icon: str = defaults.TRANSITION
    "Icon/emoji to use when decorating this transition visually."

    # noinspection PyUnresolvedReferences
    @name.default
    def _default_name_is_id(self):
        """By default, `name` is set to the `id` if not passed in."""
        return self.id

    def __hash__(self):
        return hash(self.id)

    def __lt__(self, other):
        """Transitions are sorted by name."""
        return self.name < other.name

    def __repr__(self):
        if self.icon:
            return f"{self.icon} {self.name}"
        if self.name != self.id:
            return self.name
        return f"<Transition id={self.id!r}>"

    def __lshift__(self, other: Place | Color | ColorSet | AbstractSet[Color]) -> ArcPT:
        """Returns an arc from this `Transition` from the given `Place`."""

        from carladam.petrinet.arc import ArcPT, CompletedArcPT
        from carladam.petrinet.place import Place

        if isinstance(other, Color):
            other = {other: 1}
        if isinstance(other, (set, frozenset, PSet)):
            other = {color: 1 for color in other}
        if isinstance(other, dict):
            return ArcPT(src=None, dest=self, weight=other)
        if isinstance(other, Place):
            return CompletedArcPT(src=other, dest=self)
        raise TypeError("Transition cannot be connected to object.", other)

    @overload
    def __rshift__(self, other: Place) -> CompletedArcTP: ...

    @overload
    def __rshift__(self, other: Color | ColorSet | AbstractSet[Color]) -> ArcTP: ...

    def __rshift__(self, other: Place | Color | ColorSet | AbstractSet[Color]) -> ArcTP | CompletedArcTP:
        """Returns an arc from this `Transition` to the given `Place`."""

        from carladam.petrinet.arc import ArcTP, CompletedArcTP
        from carladam.petrinet.place import Place

        if isinstance(other, Color):
            other = {other: 1}
        if isinstance(other, (set, frozenset, PSet)):
            other = {color: 1 for color in other}
        if isinstance(other, dict):
            return ArcTP(src=self, dest=None, weight=other)
        if isinstance(other, Place):
            return CompletedArcTP(src=self, dest=other)
        raise TypeError("Transition cannot be connected to object.", other)



"""
CarlAdam models Petri nets in Python.

This top-level module exports commonly used classes and functions.
"""

from carladam.petrinet.arc import Annotate, TransformEach, arc, arc_path
from carladam.petrinet.color import Abstract, Color, color_eq
from carladam.petrinet.marking import Marking
from carladam.petrinet.petrinet import PetriNet
from carladam.petrinet.place import Place
from carladam.petrinet.token import Token, TokenSet, one, tokens_where
from carladam.petrinet.transition import Transition, passthrough
from carladam.util.autoname import autoname

__ = autoname
"Convenience alias for `autoname` to reduce noise in notebook cells."

__all__ = [
    "Abstract",
    "Annotate",
    "Color",
    "Marking",
    "PetriNet",
    "Place",
    "Token",
    "TokenSet",
    "TransformEach",
    "Transition",
    "__",
    "arc",
    "arc_path",
    "autoname",
    "color_eq",
    "one",
    "passthrough",
    "tokens_where",
]



from uuid import uuid4


def default_id() -> str:
    """Returns a random unique UUID string."""
    return uuid4().hex


TRANSITION = "□"
"""Decoration for Transition instances."""

PLACE = "⬭"
"""Decoration for Place instances."""

ARROW = "→"
"""Arrow decoration for Arc instances."""

EMPTY_SET = "∅"
"""Decorative representation of empty set."""

ABSTRACT_TOKEN = "⚫️"
"""Color name given to abstract tokens."""

INHIBITOR = "⭘"
"""Decoration for inhibitor arcs."""



"""Generates sequence diagrams of Petri net runs using PlantUML."""

from __future__ import annotations

from functools import lru_cache
from textwrap import wrap
from typing import Sequence

from pyrsistent import plist, pmap
from pyrsistent.typing import PList as PListType

from carladam import PetriNet, Place, Token, Transition
from carladam.petrinet.marking import PMarking, pmarking


def token_held_by_place_repr(token: Token) -> str:
    if token.name == token.id:
        return repr(token.color)
    else:
        return f"{token.color!r} {token.name}"


def wrapped(s: str, width: int = 12) -> str:
    return "\\n".join(wrap(s, width=width))


def plantuml_sequence_diagram(
    net: PetriNet,
    initial_marking: PMarking = pmap(),
    transitions: Sequence[Transition] = plist(),
) -> str:
    """Return a PlantUML representation of a run given a net, initial marking, and sequence of transitions."""
    return _plantuml_sequence_diagram(
        net=net,
        initial_marking=pmarking(initial_marking),
        transitions=plist(transitions),
    )


@lru_cache
def _plantuml_sequence_diagram(
    net: PetriNet,
    initial_marking: PMarking = pmap(),
    transitions: PListType[Transition] = plist(),
) -> str:
    places_added: set[Place] = set()
    transitions_added: set[Transition] = set()

    participants: list[str] = []
    events: list[str] = []

    header = ["!pragma teoz true"]

    def add_place(p: Place):
        if p in places_added:
            return
        print("add_place", p)
        places_added.add(p)
        participants.append(f'queue "{wrapped(p.name)}" as p_{p.id}')

    def add_transition(t: Transition):
        if t in transitions_added:
            return
        print("add_transition", t)
        transitions_added.add(t)
        participants.append(f'participant "{wrapped(t.name)}" as t_{t.id}')

    def add_marking_event(marking: PMarking):
        print("add_marking_event")
        for place, tokens in marking.items():
            if tokens:
                add_place(place)
        marking_repr = {
            place: "".join(token_held_by_place_repr(token) for token in sorted(tokens))
            for place, tokens in marking.items()
        }
        marking_notes = "\n& ".join(
            f"note over p_{place.id}\n{marking_repr[place]}\nend note" for place, tokens in marking.items() if tokens
        )
        events.append(marking_notes)

    def add_transition_event(transition: Transition):
        print("add_transition_event", transition)
        add_transition(transition)
        transition_arcs = "".join(
            [
                "\n& ".join(f"p_{arc.src.id} x-> t_{transition.id}" for arc in net.node_inputs.get(transition, ())),
                f"\nactivate t_{transition.id} #skyblue\n",
                "\n& ".join(f"t_{transition.id} ->o p_{arc.dest.id}" for arc in net.node_outputs.get(transition, ())),
                f"\ndeactivate t_{transition.id}\n",
            ]
        )
        events.append(transition_arcs)

    current_marking = initial_marking
    add_marking_event(current_marking)
    for current_transition in transitions:
        add_transition_event(current_transition)
        current_marking = net.marking_after_transition(current_marking, current_transition)
        add_marking_event(current_marking)

    return "\n".join(header + participants + participants + events)



"""Interface to the Kroki and Niolesk services."""

import base64
import os
import zlib
from functools import lru_cache

DEFAULT_KROKI_BASE_URL = "https://kroki.io/"
DEFAULT_NIOLESK_BASE_URL = "https://niolesk.top/"
DEFAULT_NIOLESK_KROKI_URL = DEFAULT_KROKI_BASE_URL

KROKI_BASE_URL = os.getenv("CARLADAM_KROKI_BASE_URL", DEFAULT_KROKI_BASE_URL)
NIOLESK_BASE_URL = os.getenv("CARLADAM_NIOLESK_BASE_URL", DEFAULT_NIOLESK_BASE_URL)
NIOLESK_KROKI_URL = os.getenv("CARLADAM_NIOLESK_KROKI_URL", DEFAULT_NIOLESK_KROKI_URL)

KROKI_MARKDOWN_TEMPLATE = """\
![Diagram]({image_url})

[Edit this diagram]({niolesk_url}) [Pop out]({image_url})
"""


@lru_cache
def kroki_encoded(s: str) -> str:
    data = s.encode("utf8")
    return base64.urlsafe_b64encode(zlib.compress(data, 9)).decode("ascii")


@lru_cache
def kroki_image_url(
    diagram_source: str,
    diagram_type: str = "plantuml",
    image_format: str = "svg",
    base_url: str = KROKI_BASE_URL,
) -> str:
    encoded_source = kroki_encoded(diagram_source)
    return f"{base_url}{diagram_type}/{image_format}/{encoded_source}"


@lru_cache
def niolesk_edit_url(image_url: str) -> str:
    return f"{NIOLESK_BASE_URL}#{image_url}"



from django.urls import path

from carladam.django.petrinet_simulator import views

urlpatterns = [
    path("", views.index, name="index"),
    path("<str:net_name>/", views.simulator, name="simulator"),
]



from __future__ import annotations

from collections import Counter
from typing import Mapping, Sequence, TYPE_CHECKING

import attrs
from pyrsistent import plist, pmap, pset
from pyrsistent.typing import PList

from carladam.petrinet.arc import CompletedArcPT, CompletedArcTP
from carladam.petrinet.color import ColorSet
from carladam.petrinet.effects import Consume, Effect, Input, Output, Produce
from carladam.petrinet.errors import (
    ArcGuardRaisesException,
    ArcGuardReturnsFalse,
    PetriNetTransitionFunctionOutputHasOverlappingColorsets,
    TransitionGuardRaisesException,
    TransitionGuardReturnsFalse,
    TransitionHasNoArcs,
    TransitionNotEnabled,
)
from carladam.petrinet.marking import PMarking
from carladam.petrinet.token import TokenSet
from carladam.petrinet.transition import Transition

if TYPE_CHECKING:
    from carladam.petrinet.petrinet import PetriNet  # pragma: nocover


@attrs.define
class Occurrence:
    net: PetriNet
    marking: PMarking
    transition: Transition

    def is_enabled(self) -> bool:
        try:
            self.check_enabled()
        except TransitionNotEnabled:
            return False
        return True

    def check_enabled(self):
        self._check_transition_connectivity()
        for arc in self.input_arcs():
            tokens = self.marking.get(arc.src, set())
            self._check_arc_guard(arc, tokens)
        self._check_transition_guard()

    def _check_transition_connectivity(self):
        """Checks that the transition is connected to at least one arc."""
        input_arcs = self.net.node_inputs.get(self.transition)
        output_arcs = self.net.node_outputs.get(self.transition)
        if not input_arcs and not output_arcs:
            raise TransitionHasNoArcs()

    def _check_arc_guard(self, arc, tokens):
        """
        Checks that the arc guard passes.

        [NOTE]: This checks tokens that will be consumed from a place,
         *before* an arc's `transform` method is called.
         See also https://github.com/matthewryanscott/CarlAdam/issues/14
        """
        try:
            arc_guard_result = arc.guard(arc, tokens)
        except Exception as e:
            raise ArcGuardRaisesException(arc, tokens) from e
        if not arc_guard_result:
            raise ArcGuardReturnsFalse()

    def _check_transition_guard(self):
        """Checks that the transition guard passes."""
        try:
            guard_result = self.transition.guard(self.transition_inputs())
        except Exception as e:
            raise TransitionGuardRaisesException() from e
        if not guard_result:
            raise TransitionGuardReturnsFalse()

    def input_arcs(self) -> Sequence[CompletedArcPT]:
        return self.net.node_inputs.get(self.transition, ())

    def transition_inputs(self):
        inputs = set()
        for arc in self.input_arcs():
            place = arc.src
            new_place_tokens = set(self.marking.get(place) or set())
            colors_left = dict(arc.weight)
            for token in new_place_tokens.copy():
                if colors_left.get(token.color, 0):
                    inputs.add(token)
                    new_place_tokens.remove(token)
                    colors_left[token.color] -= 1
        return pset(inputs)

    def effects(self) -> PList[Effect]:
        self.check_enabled()
        effects = []
        # Consume inputs.
        inputs = set()
        for arc in self.input_arcs():
            place = arc.src
            colors_left = dict(arc.weight)
            inputs_to_add = set()
            tokens = self.marking.get(place, ())
            for token in tokens:
                if colors_left.get(token.color, 0):
                    inputs_to_add.add(token)
                    effects.append(Consume(arc=arc, token=token))
                    colors_left[token.color] -= 1
            if callable(arc.transform):
                inputs_to_add = arc.transform(inputs_to_add)
            effects.extend(Input(arc=arc, token=token) for token in inputs_to_add)
            inputs.update(inputs_to_add)
        # Calculate outputs.
        output_tokensets = pset(pset(tokenset) for tokenset in self.transition.fn(pset(inputs)))
        # Ensure each output's colorset is unique amongst all outputs.
        # This is to avoid ambiguity when placing tokens into destinations.
        output_tokensets_by_colorset: Mapping[ColorSet, TokenSet] = {
            pmap(Counter(token.color for token in tokenset)): tokenset for tokenset in output_tokensets
        }
        if len(output_tokensets_by_colorset) < len(output_tokensets):
            raise PetriNetTransitionFunctionOutputHasOverlappingColorsets()
        # Produce outputs.
        for arc in self.output_arcs():
            outputs_for_place = output_tokensets_by_colorset[arc.weight]
            effects.extend(Output(arc=arc, token=token) for token in outputs_for_place)
            if callable(arc.transform):
                outputs_for_place = arc.transform(outputs_for_place)
            effects.extend(Produce(arc=arc, token=token) for token in outputs_for_place)
        return plist(effects)

    def output_arcs(self) -> Sequence[CompletedArcTP]:
        return self.net.node_outputs.get(self.transition, ())



import inspect
import typing
from collections import Counter
from typing import Type

from carladam.petrinet.place import Place
from carladam.petrinet.token import Token
from carladam.petrinet.transition import Transition

AutoNameableTypes = (Place, Token, Transition)
"Types that can be automatically given a name."

AUTONAME_TYPE_COUNTER: typing.Counter[Type] = Counter()


def capitalize(key: str) -> str:
    """Default function used to autoname."""
    return key.replace("_", " ").capitalize()


AUTONAME_ID_SHORTENED = {
    "Place": "P",
    "Transition": "T",
}


def autoname(*objects, set_id=True, autoname_fn=capitalize):
    """
    Auto-name CarlAdam objects wherever possible, based on the calling stack as needed.

    Examples::

        >>> from carladam import *
        >>> my_place = Place()
        >>> my_place  #doctest: +ELLIPSIS
        ⬭ ...
        >>> autoname(my_place)
        ⬭ My place
        >>> my_place
        ⬭ My place

        >>> autoname(my_transition := Transition())
        □ My transition

        >>> net = PetriNet.new(autoname(
        ...     p0 := Place(),
        ...     t0 := Transition(),
        ...     p1 := Place(),
        ...     a0 := p0 >> t0,
        ...     a1 := t0 >> p1,
        ... ))
        >>> a0
        ⬭ P0 → □ T0
    """
    if len(objects) == 1 and isinstance(single_class := objects[0], type):
        # The namespace is a class's attributes.
        # The objects are all of those class's attribute values.
        d = single_class.__dict__
        items = d.items()
        objects = d.values()
        return_value = single_class
    else:
        # The namespace is the calling stack frame.
        # The object(s) are those which were passed in to autoname.
        items = inspect.stack()[1].frame.f_locals.items()
        return_value = objects[0] if len(objects) == 1 else objects

    # Autoname objects that are in the namespace.
    for key, value in items:
        if isinstance(value, type):
            value = autoname(value)
        if key == "_":
            # Always ignore _.
            continue
        if not isinstance(value, AutoNameableTypes):
            # Not a nameable object.
            continue
        if value.name != value.id:
            # Object already given a name.
            continue
        if value not in objects:
            # Did not ask for this object to be autonamed.
            continue

        # Set the new name.
        value.name = autoname_fn(key)

        if set_id:
            # Ensure that other values of this type having this name are given the same ID.
            value_type = type(value)
            value_type_name = value_type.__name__
            value_type_name = AUTONAME_ID_SHORTENED.get(value_type_name, value_type_name)
            AUTONAME_TYPE_COUNTER.update({value_type})
            value.id = f"{value_type_name}{AUTONAME_TYPE_COUNTER[value_type]}{key}"

    return return_value



from __future__ import annotations

import json
from functools import lru_cache
from textwrap import dedent
from typing import Sequence
from urllib.parse import quote

import httpx
from django import template
from django.utils.safestring import SafeString
from markdown import markdown as _markdown
from pyrsistent import PMap, thaw

from carladam import Marking, PetriNet, Transition
from carladam.diagram.digraph import graphviz_digraph
from carladam.diagram.kroki import NIOLESK_KROKI_URL, kroki_image_url, niolesk_edit_url
from carladam.diagram.sequence import plantuml_sequence_diagram
from carladam.petrinet.marking import PMarking, pmarking
from carladam.petrinet.types import PetriNetNode

register = template.Library()


@register.filter
def markdown(source: str):
    return SafeString(_markdown(source))


@register.filter
def net_title(net: PetriNet):
    return (net.__doc__ or "").lstrip().split("\n", 1)[0]


@register.filter
def net_description(net: PetriNet):
    _, *rest = (net.__doc__ or "").lstrip().split("\n", 1)
    return dedent(rest[0]).lstrip().rstrip() if rest else ""


@register.simple_tag
def net_graph_image_data(
    net: PetriNet,
    initial_marking: PMarking,
    transition_ids_str: str,
    current_marking: PMarking,
    legend: bool = True,
    rotate: bool = False,
) -> str:
    def transition_url_fn(t: Transition) -> str:
        return (
            f"?initial_marking={marking_encoded(initial_marking)}"
            f"&transitions={transition_ids_str},{t.id}"
            f"&rotate={int(rotate or 0)}"
        )

    diagram_source = graphviz_digraph(
        net,
        current_marking,
        legend=legend,
        transition_url_fn=transition_url_fn,
        rotate=rotate,
    )
    image_url = kroki_image_url(diagram_source=diagram_source, diagram_type="graphviz", image_format="svg")
    image_source = _image_source(image_url)
    return SafeString(image_source)


@register.simple_tag
def net_graph_niolesk_url(
    net: PetriNet,
    initial_marking: PMarking,
    transition_ids_str: str,
    current_marking: PMarking,
    legend: bool = True,
    rotate: bool = False,
) -> str:
    def transition_url_fn(t: Transition) -> str:
        return (
            f"?initial_marking={marking_encoded(initial_marking)}"
            f"&transitions={transition_ids_str},{t.id}"
            f"&rotate={int(rotate or 0)}"
        )

    diagram_source = graphviz_digraph(
        net,
        current_marking,
        legend=legend,
        transition_url_fn=transition_url_fn,
        rotate=rotate,
    )
    image_url = kroki_image_url(
        diagram_source=diagram_source,
        diagram_type="graphviz",
        image_format="svg",
        base_url=NIOLESK_KROKI_URL,
    )
    return niolesk_edit_url(image_url=image_url)


@register.simple_tag
def net_run_sequence_diagram_image_data(
    net: PetriNet,
    initial_marking: PMarking,
    transitions: Sequence[Transition],
) -> str:
    diagram_source = plantuml_sequence_diagram(net=net, initial_marking=initial_marking, transitions=transitions)
    image_url = kroki_image_url(diagram_source=diagram_source, diagram_type="plantuml", image_format="svg")
    image_source = _image_source(image_url)
    return SafeString(image_source)


@register.simple_tag
def net_run_sequence_diagram_niolesk_url(
    net: PetriNet,
    initial_marking: PMarking,
    transitions: Sequence[Transition],
) -> str:
    diagram_source = plantuml_sequence_diagram(net=net, initial_marking=initial_marking, transitions=transitions)
    image_url = kroki_image_url(
        diagram_source=diagram_source,
        diagram_type="plantuml",
        image_format="svg",
        base_url=NIOLESK_KROKI_URL,
    )
    return niolesk_edit_url(image_url=image_url)


@register.filter
def marking_encoded(marking: Marking):
    if not isinstance(marking, PMap):
        marking = pmarking(marking)
    return _marking_encoded(marking)


@lru_cache
def _marking_encoded(marking: PMarking):
    marking_json = {
        place.id: [
            {
                "id": token.id,
                "color": token.color.label,
                "data": token.data,
            }
            for token in tokens
        ]
        for place, tokens in marking.items()
    }
    return quote(json.dumps(thaw(marking_json)))


@register.filter
def net_subnet(net: PetriNet, node: PetriNetNode):
    return net.subnet(node)


@lru_cache
def _image_source(image_url: str) -> str:
    return httpx.get(image_url).text



from typing import Callable, Dict

import networkx as nx
from numpy.typing import ArrayLike

from carladam.petrinet.petrinet import PetriNet
from carladam.petrinet.types import PetriNetNode

NodeCoordinates = Dict[PetriNetNode, ArrayLike]
LayoutFunction = Callable[..., NodeCoordinates]


def auto_layout_nodes(net: PetriNet, algorithm: LayoutFunction = nx.spring_layout, **kwargs) -> NodeCoordinates:
    """Automatically layout node geometry based on a layout function."""
    graph = networkx_graph(net)
    return algorithm(graph, **kwargs)


def networkx_graph(net: PetriNet) -> nx.Graph:
    """Transforms a `PetriNet` into a `Graph` suitable for processing with `networkx`."""
    graph = nx.Graph()
    graph.add_nodes_from(net.places)
    graph.add_nodes_from(net.transitions)
    for arc in net.arcs:
        graph.add_edge(arc.src, arc.dest)
    return graph



"""Generates directed graph diagrams of Petri nets w/ markings using GraphViz."""

from __future__ import annotations

from collections import defaultdict
from functools import lru_cache
from textwrap import dedent, wrap
from typing import Callable, DefaultDict, FrozenSet, Optional, Set, Tuple

from jinja2 import Template
from pyrsistent import pmap

from carladam import Place, Transition
from carladam.petrinet import defaults
from carladam.petrinet.arc import ArcPT, CompletedArcPT
from carladam.petrinet.color import colorset_string
from carladam.petrinet.marking import PMarking
from carladam.petrinet.petrinet import PetriNet
from carladam.petrinet.token import Token
from carladam.petrinet.types import CompletedArc

PLACE_ATTRIBUTES = "[shape=oval]"
TRANSITION_ATTRIBUTES = "[shape=box]"

PLACE_WITH_TOKEN_ATTRIBUTES = "[style=filled,fillcolor=ivory]"

ENABLED_TRANSITION_ATTRIBUTES = "[style=filled,color=blue,fillcolor=skyblue,fontcolor=navyblue]"
DISABLED_TRANSITION_ATTRIBUTES = "[style=filled,color=navyblue,fillcolor=aliceblue,fontcolor=navyblue]"

DISABLED_ARC_PT_ATTRIBUTES = "[color=gray]"
DISABLED_ARC_TP_ATTRIBUTES = "[color=skyblue]"
ENABLED_ARC_PT_ATTRIBUTES = "[color=black,penwidth=2]"
ENABLED_ARC_TP_ATTRIBUTES = "[color=blue,penwidth=2]"

DEFAULT_LEGEND_WIDTH = 25


def token_held_by_place_repr(token: Token) -> str:
    if token.name == token.id:
        return repr(token.color)
    else:
        return f"{token.color!r} {token.name}"


def wrapped(s: str, width: int = 12) -> str:
    return "\n".join(wrap(s, width=width))


EdgeArcWeight = FrozenSet
EdgeArcAnnotation = Optional[str]
EdgeArcsKey = Tuple[EdgeArcWeight, EdgeArcAnnotation, Place, Transition]
EdgeArcs = DefaultDict[EdgeArcsKey, Set[CompletedArc]]


def always_none(transition):
    return None


@lru_cache
def graphviz_digraph(
    net: PetriNet,
    marking: PMarking = pmap(),
    /,
    legend: bool = False,
    rotate: bool = False,
    legend_width: int = DEFAULT_LEGEND_WIDTH,
    transition_url_fn: Callable[[Transition], str] | None = None,
) -> str:
    """Return a GraphViz representation of a `PetriNet`."""
    transition_url_fn = transition_url_fn or always_none

    marking_repr = {
        place: "".join(token_held_by_place_repr(token) for token in sorted(tokens)) for place, tokens in marking.items()
    }

    arcs = set(net.arcs)
    double_arcs = set()
    edge_arcs: EdgeArcs = defaultdict(set)
    for arc in arcs:
        if isinstance(arc, ArcPT):
            place = arc.src
            transition = arc.dest
        else:
            place = arc.dest
            transition = arc.src
        edge_arcs[(frozenset(arc.weight.items()), arc.annotation, place, transition)].add(arc)
    for pair in edge_arcs.values():
        if len(pair) != 2:
            continue
        for arc in pair:
            arcs.remove(arc)
            if isinstance(arc, CompletedArcPT):
                double_arcs.add(arc)

    sorted_places = list(sorted(net.places))
    enabled_transitions = set(net.enabled_transitions(marking))
    disabled_transitions = net.transitions.difference(enabled_transitions)
    sorted_enabled_transitions = list(sorted(enabled_transitions))
    sorted_disabled_transitions = list(sorted(disabled_transitions))
    graphviz_source = GRAPHVIZ_DIGRAPH_TEMPLATE.render(
        DISABLED_ARC_PT_ATTRIBUTES=DISABLED_ARC_PT_ATTRIBUTES,
        DISABLED_ARC_TP_ATTRIBUTES=DISABLED_ARC_TP_ATTRIBUTES,
        DISABLED_TRANSITION_ATTRIBUTES=DISABLED_TRANSITION_ATTRIBUTES,
        EMPTY_SET=defaults.EMPTY_SET,
        ENABLED_ARC_PT_ATTRIBUTES=ENABLED_ARC_PT_ATTRIBUTES,
        ENABLED_ARC_TP_ATTRIBUTES=ENABLED_ARC_TP_ATTRIBUTES,
        ENABLED_TRANSITION_ATTRIBUTES=ENABLED_TRANSITION_ATTRIBUTES,
        PLACE_ATTRIBUTES=PLACE_ATTRIBUTES,
        PLACE_WITH_TOKEN_ATTRIBUTES=PLACE_WITH_TOKEN_ATTRIBUTES,
        TRANSITION_ATTRIBUTES=TRANSITION_ATTRIBUTES,
        arcs=sorted(arcs),
        clusters=getattr(net.structure, "clusters", {}),
        colorset_string=colorset_string,
        disabled_transitions=sorted_disabled_transitions,
        double_arcs=sorted(double_arcs),
        enabled_transitions=sorted_enabled_transitions,
        enumerate=enumerate,
        legend=legend,
        legend_width=legend_width,
        marking=marking,
        marking_repr=marking_repr,
        places=sorted_places,
        rotate=rotate,
        sorted=sorted,
        str=str,
        transition_url_fn=transition_url_fn,
        wrap=wrap,
        wrapped=wrapped,
    )
    return dedent(graphviz_source)


# noinspection HtmlDeprecatedAttribute
GRAPHVIZ_DIGRAPH_TEMPLATE = Template(
    # language=jinja2
    """
    digraph {
        layout=dot
        overlap=false
        {% if rotate %}rankdir=LR{% endif %}
        {
            node [fontname="-apple-system,BlinkMacSystemFont,sans-serif"]
            edge [fontname="-apple-system,BlinkMacSystemFont,sans-serif"]
            graph [color=lightgray,fontcolor=lightgray,fontname="-apple-system,BlinkMacSystemFont,sans-serif"]

            node {{ PLACE_ATTRIBUTES }}
            {% for place in places %}
                n_{{ place.id }}
                [label="{{ wrapped(place.name) }}{% if marking_repr.get(place) %}\n{{ marking_repr[place] }}{% endif %}"]
                {% if marking[place] -%}
                    {{ PLACE_WITH_TOKEN_ATTRIBUTES }}
                {%- endif %}
            {% endfor %}

            node {{ TRANSITION_ATTRIBUTES }}
            {% for transition in enabled_transitions %}
                n_{{ transition.id }}
                {{ ENABLED_TRANSITION_ATTRIBUTES }}
                [label="{{ wrapped(transition.name) }}{% if transition.annotation %}\n\n{{ transition.annotation }}{% endif %}"]
                {% set transition_url = transition_url_fn(transition) %}
                {% if transition_url %}
                    [URL="{{ transition_url }}"]
                {% endif %}
            {% endfor %}
            {% for transition in disabled_transitions %}
                n_{{ transition.id }}
                {{ DISABLED_TRANSITION_ATTRIBUTES }}
                [label="{{ wrapped(transition.name) }}{% if transition.annotation %}\n\n{{ transition.annotation }}{% endif %}"]
            {% endfor %}

            {% for arc in arcs %}
                n_{{ arc.src.id }}->n_{{ arc.dest.id }}
                [label="{{ colorset_string(arc.weight) }}{% if arc.annotation %} {{ wrapped(arc.annotation) }}{% endif %}"]
                {% if arc.src in places -%}
                    {%- if arc.dest in enabled_transitions -%}
                        {{ ENABLED_ARC_PT_ATTRIBUTES }}
                    {%- else -%}
                        {{ DISABLED_ARC_PT_ATTRIBUTES }}
                    {%- endif -%}
                {%- else -%}
                    {%- if arc.src in enabled_transitions -%}
                        {{ ENABLED_ARC_TP_ATTRIBUTES }}
                    {%- else -%}
                        {{ DISABLED_ARC_TP_ATTRIBUTES }}
                    {%- endif -%}
                {%- endif %}
            {% endfor %}

            {% for arc in double_arcs %}
                n_{{ arc.src.id }}->n_{{ arc.dest.id }}
                [label="{{ colorset_string(arc.weight) }}{% if arc.annotation %} {{ wrapped(arc.annotation) }}{% endif %}"]
                [dir=both]
                {%- if arc.dest in enabled_transitions -%}
                    {{ ENABLED_ARC_PT_ATTRIBUTES }}
                {%- else -%}
                    {{ DISABLED_ARC_PT_ATTRIBUTES }}
                {%- endif -%}
            {% endfor %}

            {% for index, (name, cluster) in enumerate(clusters.items()) %}
                subgraph cluster_{{ index }} {
                    label="{{ name }}"
                    {% for item in cluster %}
                        n_{{ item.id }}
                    {% endfor %}
                }
            {% endfor %}
        }
        {% if marking and legend %}
            {
                rank=max
                MarkingLegend [shape=none, margin=0, label=<
                    <table border="0" cellborder="0" cellspacing="0" cellpadding="4">
                        <tr>
                            <td colspan="2"><b>Marking</b></td>
                        </tr>
                        {% for place, tokens in sorted(marking.items()) %}
                            {% if place in places and tokens %}
                                <tr>
                                    <td align="left" valign="top">{{ place.name }}:</td>
                                    <td align="left" valign="top">{% for token in tokens -%}
                                        {%- for line in wrap(str(token), legend_width) -%}
                                            {{ line }}
                                            {%- if not loop.last %}<br/>  {% endif -%}
                                        {%- endfor -%}
                                        {%- if not loop.last %}<br/>{% endif -%}
                                    {%- endfor %}</td>
                                </tr>
                            {% endif %}
                        {% endfor %}
                    </table>
                >]
            }
        {% endif %}
    }
    """  # noqa: E501 line too long
)



from typing import Iterable, Union

from carladam.petrinet.arc import ArcPT, ArcTP, CompletedArcPT, CompletedArcTP
from carladam.petrinet.place import Place
from carladam.petrinet.transition import Transition

Arc = Union[ArcPT, ArcTP]
"Union of arc types, used for type annotation."

CompletedArc = Union[CompletedArcPT, CompletedArcTP]
"Union of completed arc types, used for type annotation."

ArcTypes = (ArcPT, ArcTP)
"Tuple of arc types, used for `isinstance` checking."

CompletedArcTypes = (CompletedArcPT, CompletedArcTP)
"Tuple of completed arc types, used for `isinstance` checking."

PetriNetMember = Union[Place, Transition, Arc]
"Object which can be a member of a `PetriNet`."

PetriNetMemberSet = Iterable[PetriNetMember]
"Iterable of objects which can be members of a `PetriNet`."

PetriNetMemberOrSet = Union[PetriNetMember, PetriNetMemberSet]
"Object or iterable of objects which can be members of a `PetriNet`."

PetriNetNode = Union[Place, Transition]
"Object that is a node in a `PetriNet`."



from __future__ import annotations

from typing import Sequence

import attrs
from pyrsistent import pset

from carladam.petrinet.arc import CompletedArcPT, CompletedArcTP
from carladam.petrinet.marking import PMarking
from carladam.petrinet.token import Token


@attrs.define
class Consume:
    """
    A token was consumed from a place by an arc.

    The token will always represent a token that was present in a place prior to a transition occurrence.
    """

    arc: CompletedArcPT
    token: Token

    def apply_to_marking(self, marking: PMarking) -> PMarking:
        new_place_tokens = marking.get(self.arc.src, pset()).remove(self.token)
        if new_place_tokens:
            return marking.set(self.arc.src, pset(new_place_tokens))
        return marking.remove(self.arc.src)


@attrs.define
class Input:
    """
    A token was passed to a transition as an input.

    The token may be one that was consumed from a place, or one that was generated by an arc's `transform` function.
    """

    arc: CompletedArcPT
    token: Token

    def apply_to_marking(self, marking: PMarking) -> PMarking:
        return marking


@attrs.define
class Output:
    """
    A token was returned by a transition as an output.

    The token will always represent a token that was generated by the transition's `fn` function.
    """

    arc: CompletedArcTP
    token: Token

    def apply_to_marking(self, marking: PMarking) -> PMarking:
        return marking


@attrs.define
class Produce:
    """
    A token was produced to a place by an arc.

    The token may be one that was generated by a transition's `fn` function,
    or it may be one that was generated by an arc's `transform` function.
    """

    arc: CompletedArcTP
    token: Token

    def apply_to_marking(self, marking: PMarking) -> PMarking:
        new_place_tokens = marking.get(self.arc.dest, pset()).add(self.token)
        return marking.set(self.arc.dest, pset(new_place_tokens))


Effect = Consume | Input | Output | Produce


def apply_effects_to_marking(marking: PMarking, effects: Sequence[Effect]) -> PMarking:
    for effect in effects:
        marking = effect.apply_to_marking(marking)
    return marking



"""
An Arc is a directed connection from `Place` → `Transition` or `Transition` → `Place`.

Note: The classes in this module contain a lot of duplication.
These will be cleaned up in a future effort that will remove the use of `attrs` which is subclass-unfriendly.
"""

from __future__ import annotations

from collections import Counter
from functools import lru_cache
from typing import AbstractSet, Callable, Iterator, TYPE_CHECKING, overload

from attr import Factory, define, field
from attr.validators import instance_of, optional
from pyrsistent import pmap

from carladam.petrinet import defaults, errors
from carladam.petrinet.color import Abstract, Color, ColorSet, colorset_string
from carladam.petrinet.defaults import INHIBITOR
from carladam.petrinet.place import Place
from carladam.petrinet.token import Token, TokenSet
from carladam.petrinet.transition import Transition

if TYPE_CHECKING:  # pragma: nocover
    from carladam.petrinet.types import Arc


@overload
def arc(
    src: Place,
    dest: Transition,
    weight: Color | ColorSet = None,
    *,
    annotation: str = None,
    transform: Callable = None,
    guard: Callable = None,
) -> CompletedArcPT: ...


@overload
def arc(
    src: Transition,
    dest: Place,
    weight: Color | ColorSet = None,
    *,
    annotation: str = None,
    transform: Callable = None,
    guard: Callable = None,
) -> CompletedArcTP: ...


def arc(src, dest, weight=None, **kwargs):
    if isinstance(weight, Color):
        weight = {weight: 1}
    elif weight is None:
        weight = {Abstract: 1}
    if isinstance(src, Place) and isinstance(dest, Transition):
        return CompletedArcPT(src, dest, weight, **kwargs)
    if isinstance(src, Transition) and isinstance(dest, Place):
        return CompletedArcTP(src, dest, weight, **kwargs)
    raise TypeError("Arcs must be from Place to Transition or Transition to Place.")


def arc_path(*nodes):
    """
    Link places and transitions together in series using `arc`.

    Example:

        >>> from carladam import Place, Transition, arc, arc_path
        >>> p1, p2, p3 = Place(), Place(), Place()
        >>> t1, t2 = Transition(), Transition()
        >>> list(arc_path(p1, t1, p2, t2, p3)) == [arc(p1, t1), arc(t1, p2), arc(p2, t2), arc(t2, p3)]
        True
    """
    for src, dest in zip(nodes, nodes[1:]):
        yield arc(src, dest)


def inhibitor_arc(src: Place, dest: Transition, *args, **kwargs):
    return arc(src, dest, *args, annotation=INHIBITOR, **kwargs, guard=inhibit)


def __arc_hash__(self):
    """Common implementation of `__hash__` for all Arc types."""
    return hash((self.src, self.dest, frozenset(self.weight.items())))


def __arc_lt__(self, other):
    """Common implementation of `__lt__` for all Arc types."""
    return (self.src.name, self.dest.name) < (other.src.name, other.dest.name)


def __arc_repr__(self: Arc):
    """Common implementation of `__repr__` for all Arc types."""
    annotation = f" {self.annotation}" if self.annotation else ""
    colorset = f" {s}" if (s := colorset_string(self.weight)) else ""
    return f"{self.src!r}{colorset}{annotation} {defaults.ARROW} {self.dest!r}"


@lru_cache
def default_arc_weight() -> ColorSet:
    """By default, arcs transmit a single Abtract token."""
    return pmap({Abstract: 1})


def weights_are_satisfied(arc: CompletedArcPT, tokens: AbstractSet[Token]) -> bool:
    colors: ColorSet = Counter(token.color for token in tokens)
    # Do the tokens have all the colors specified by the arc weight?
    if frozenset(arc.weight) - frozenset(colors):
        return False
    # Do the tokens satisfy all of the quantities specified by the arc weight?
    return all(arc.weight[color] <= count for color, count in colors.items())


def inhibit(arc: CompletedArcPT, tokens: AbstractSet[Token]) -> bool:
    return not tokens


@define
class ArcPT:
    """An arc from `Place` → `Transition`."""

    src: Place | None = field(default=None, validator=optional(instance_of(Place)))
    """The place where the transition will consume token(s) from."""

    dest: Transition | None = field(default=None, validator=optional(instance_of(Transition)))
    """The transition that will consume token(s) from the place."""

    weight: ColorSet = Factory(default_arc_weight)
    """The type and amount of tokens that will be consumed."""

    annotation: str | None = None
    """Human-friendly annotation to display with this arc when representing it."""

    # noinspection PyUnresolvedReferences
    transform: Callable | None = None
    """Function that will transform each token consumed."""

    guard: Callable = weights_are_satisfied
    """Function that returns True if the input tokens pass criteria."""

    completed: bool = False
    """Whether this arc has both a src and dest given."""

    def __lshift__(self, other: Place | Annotate | TransformEach) -> ArcPT:
        if isinstance(other, (Annotate, TransformEach)):
            return other.apply_to_arc(self)
        if self.dest is None:
            raise errors.PetriNetArcIncomplete("Cannot << to an arc not having a dest")
        return CompletedArcPT(
            src=other,
            dest=self.dest,
            weight=self.weight,  # type: ignore
            annotation=self.annotation,
            transform=self.transform,
        )

    def __rshift__(self, other: Transition | Annotate | TransformEach) -> ArcPT:
        if isinstance(other, (Annotate, TransformEach)):
            return other.apply_to_arc(self)
        if self.src is None:
            raise errors.PetriNetArcIncomplete("Cannot >> from an arc not having a src")
        return CompletedArcPT(
            src=self.src,
            dest=other,
            weight=self.weight,  # type: ignore
            annotation=self.annotation,
            transform=self.transform,
        )

    def __call__(self, *args, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        return self


@define
class CompletedArcPT(ArcPT):
    """A completely-specified ArcPT."""

    src: Place = field(validator=instance_of(Place))
    dest: Transition = field(validator=instance_of(Transition))
    weight: ColorSet = Factory(default_arc_weight)
    annotation: str | None = None
    # noinspection PyUnresolvedReferences
    transform: Callable | None = None
    guard: Callable = weights_are_satisfied
    completed: bool = True

    __hash__ = __arc_hash__
    __lt__ = __arc_lt__
    __repr__ = __arc_repr__


@define
class ArcTP:
    """An arc from `Transition` → `Place`."""

    src: Transition | None = field(default=None, validator=optional(instance_of(Transition)))
    """The transition that will produce token(s) to the place."""

    dest: Place | None = field(default=None, validator=optional(instance_of(Place)))
    """The place where the transition will produce token(s) to."""

    weight: ColorSet = field(factory=default_arc_weight, converter=pmap)
    """The type and amount of tokens that will be consumed."""

    annotation: str | None = None
    """Human-friendly annotation to display with this arc when representing it."""

    # noinspection PyUnresolvedReferences
    transform: Callable | None = None
    """Function that will transform each token consumed."""

    completed: bool = False
    """Whether this arc has both a src and dest given."""

    def __lshift__(self, other: Transition | Annotate | TransformEach) -> ArcTP:
        if isinstance(other, (Annotate, TransformEach)):
            return other.apply_to_arc(self)
        if self.dest is None:
            raise errors.PetriNetArcIncomplete("Cannot << to an arc not having a dest")
        return CompletedArcTP(
            src=other,
            dest=self.dest,
            weight=self.weight,  # type: ignore
            annotation=self.annotation,
            transform=self.transform,
        )

    def __rshift__(self, other: Place | Annotate | TransformEach) -> ArcTP:
        if isinstance(other, (Annotate, TransformEach)):
            return other.apply_to_arc(self)
        if self.src is None:
            raise errors.PetriNetArcIncomplete("Cannot >> from an arc not having a src")
        return CompletedArcTP(
            src=self.src,
            dest=other,
            weight=self.weight,  # type: ignore
            annotation=self.annotation,
            transform=self.transform,
        )


@define
class CompletedArcTP(ArcTP):
    """A completely-specified ArcTP."""

    src: Transition = field(validator=instance_of(Transition))
    dest: Place = field(validator=instance_of(Place))
    weight: ColorSet = field(factory=default_arc_weight, converter=pmap)
    annotation: str | None = None
    # noinspection PyUnresolvedReferences
    transform: Callable | None = None
    completed: bool = True

    __hash__ = __arc_hash__
    __lt__ = __arc_lt__
    __repr__ = __arc_repr__


@define
class Annotate:
    """Decorates an `Arc` with descriptive text to show in diagrams."""

    text: str
    "Text used to describe an `Arc`."

    @overload
    def apply_to_arc(self, arc: ArcPT) -> ArcPT: ...

    @overload
    def apply_to_arc(self, arc: ArcTP) -> ArcTP: ...

    def apply_to_arc(self, arc: Arc) -> Arc:
        return type(arc)(
            src=arc.src,  # type: ignore
            dest=arc.dest,  # type: ignore
            weight=arc.weight,  # type: ignore
            annotation=self.text,
            transform=arc.transform,
        )


@define
class TransformEach:
    """Sets a function on an `Arc` that transforms each `Token` passing through it during a `Transition`."""

    fn: Callable
    "Function taking a `Token` and returning a `Token`."

    @overload
    def apply_to_arc(self, arc: ArcPT) -> ArcPT: ...

    @overload
    def apply_to_arc(self, arc: ArcTP) -> ArcTP: ...

    def apply_to_arc(self, arc: Arc) -> Arc:
        def transformed_tokens(tokens: TokenSet) -> Iterator[Token]:
            for token in tokens:
                yield self.fn(token)

        def transform(tokens: TokenSet) -> TokenSet:
            return frozenset(transformed_tokens(tokens))

        return type(arc)(
            src=arc.src,  # type: ignore
            dest=arc.dest,  # type: ignore
            weight=arc.weight,  # type: ignore
            annotation=arc.annotation,
            transform=transform,
        )



from __future__ import annotations

from collections.abc import Generator, Iterable
from functools import lru_cache
from itertools import chain
from typing import AbstractSet, Iterator, Mapping, Type, cast

from attr import Factory, define, field
from pyrsistent import PList, PMap, PSet, pmap, pset

from carladam.petrinet import errors
from carladam.petrinet.color import Abstract, Color
from carladam.petrinet.effects import apply_effects_to_marking
from carladam.petrinet.marking import Marking, PMarking, pmarking
from carladam.petrinet.occurrence import Occurrence
from carladam.petrinet.place import Place
from carladam.petrinet.transition import Transition
from carladam.petrinet.types import (
    ArcTypes,
    CompletedArc,
    CompletedArcTypes,
    PetriNetMember,
    PetriNetMemberOrSet,
    PetriNetNode,
)
from carladam.util.autoname import autoname


class PetriNetMeta(type):
    """Autoname members of the net structure class."""

    def __new__(mcs, name, bases, dct):
        cls = cast(Type["PetriNet"], super().__new__(mcs, name, bases, dct))
        autoname(cls.Structure)
        return cls


@define
class PetriNet(metaclass=PetriNetMeta):
    """A representation of a Petri net."""

    places: PSet[Place] = Factory(pset)
    "The set of ⬭ places contained within this net."

    transitions: PSet[Transition] = Factory(pset)
    "The set of □ transitions contained within this net."

    arcs: PSet[CompletedArc] = Factory(pset)
    "The set of all ⬭→□ and □→⬭ arcs contained within this net."

    node_inputs: PMap[PetriNetNode, PSet[CompletedArc]] = field(factory=pmap, repr=False)
    "Mapping used to find the inputs of a given node."

    node_outputs: PMap[PetriNetNode, PSet[CompletedArc]] = field(factory=pmap, repr=False)
    "Mapping used to find the outputs of a given node."

    structure: object = field(default=Factory(lambda self: self.Structure(), takes_self=True), eq=False, repr=False)
    """Instance of the net structure class for this net."""

    class Structure:
        """Net structure class intended to contain the definition of the Petri net."""

        clusters: Mapping[str, AbstractSet[Place | Transition]] = {}
        """Named clusters to use for visually isolating subgraphs."""

        example_markings: Mapping[str, Marking] = {}
        """Named example markings to show in a Petri net simulator."""

    @classmethod
    def new(cls, *others: PetriNetMemberOrSet):
        """Return a new instance of this class containing its net structure class and any additional members given."""
        net = cls()
        net = net.update_from_structure(net.structure)
        net = net.update(*others)
        return net

    def __contains__(self, obj: PetriNetMemberOrSet | None) -> bool:
        """Returns True if an object (or all of a set of objects) is contained somewhere in the net."""
        if obj is None:
            return False
        if isinstance(obj, Iterable):
            return all(subitem in self for subitem in obj)
        return obj in self.places or obj in self.transitions or obj in self.arcs

    def __hash__(self):
        """Instances are hashable based on their member objects."""
        return hash((self.places, self.transitions, self.node_inputs, self.node_outputs))

    def __iter__(self) -> Iterator[PetriNetMember]:
        """Iterate over all node and arc objects contained in the net."""
        return chain(self.places, self.transitions, self.arcs)

    @property
    def colors(self) -> AbstractSet[Color]:
        """Returns a list of all unique colors specified by the arcs in this net."""
        return self._colors()

    @lru_cache
    def _colors(self):
        if self.arcs:
            return pset(color for arc in self.arcs for color in arc.weight.keys())
        return pset({Abstract})

    def copy(self) -> PetriNet:
        """Returns a new net based on this one."""
        # noinspection PyArgumentList
        return self.__class__(
            places=self.places,
            transitions=self.transitions,
            arcs=self.arcs,
            node_inputs=self.node_inputs,
            node_outputs=self.node_outputs,
            structure=self.structure,
        )

    @lru_cache
    def subnet(self, node: PetriNetNode) -> PetriNet:
        """Return a new net containing only the given node, its input nodes, output nodes, and related arcs."""
        net = PetriNet.new(node)
        for arc in self.node_inputs.get(node, ()):
            net = net.update(arc.src, arc)
        for arc in self.node_outputs.get(node, ()):
            net = net.update(arc.dest, arc)
        return net

    def update(self, *others: PetriNetMemberOrSet) -> PetriNet:
        """Returns a new net based on incorporating the given object(s) into the current net."""
        new = self.copy()
        for other in others:
            if isinstance(other, (list, set, tuple, PList, PSet, Generator)):
                new = new.update(*other)
            elif isinstance(other, type):
                new = new.update(*other.__dict__.values())
            elif isinstance(other, Place) and other not in new.places:
                new.places = new.places.add(other)
            elif isinstance(other, Transition) and other not in new.transitions:
                new.transitions = new.transitions.add(other)
            elif isinstance(other, ArcTypes) and not other.completed:
                raise errors.PetriNetArcIncomplete(other.dest)
            elif isinstance(other, CompletedArcTypes) and other not in new.arcs:
                new = new.update(other.src, other.dest)
                new.arcs = new.arcs.add(other)
                if other.dest is not None:
                    new_inputs = new.node_inputs.get(other.dest, pset())
                    new_inputs = new_inputs.add(other)
                    new.node_inputs = new.node_inputs.set(other.dest, new_inputs)
                if other.src is not None:
                    new_outputs = new.node_outputs.get(other.src, pset())
                    new_outputs = new_outputs.add(other)
                    new.node_outputs = new.node_outputs.set(other.src, new_outputs)
            elif isinstance(other, PetriNet):
                new = new.update(other.places, other.transitions, other.arcs)
        new._reset_caches()
        return new

    def update_from_structure(self, structure: object) -> PetriNet:
        """Return a new net based on current net incorporating a net structure class."""
        object_items = chain(structure.__class__.__dict__.items(), structure.__dict__.items())
        object_values = (value for key, value in object_items if not key.startswith("_"))
        return self.update(object_values)

    def empty_marking(self) -> PMarking:
        """Returns an empty marking of all places in the net."""
        return pmap((place, pset()) for place in self.places)

    def enabled_transitions(self, marking: Marking) -> Iterator[Transition]:
        """Generates the transitions enabled in this net given a `Marking`."""
        for transition in self.transitions:
            if self.transition_is_enabled(marking, transition):
                yield transition

    def marking_after_transition(self, marking: Marking, transition: Transition) -> PMarking:
        """Returns the `Marking` that results from a `Transition` occuring in this net given an initial `Marking`."""
        return self._marking_after_transition(pmarking(marking), transition)

    @lru_cache
    def _marking_after_transition(self, marking: PMarking, transition: Transition) -> PMarking:
        effects = Occurrence(self, marking, transition).effects()
        return apply_effects_to_marking(marking, effects)

    def transition_is_enabled(self, marking: Marking, transition: Transition) -> bool:
        """Returns True if the given `Transition` is enabled in this net given a `Marking`."""
        return self._transition_is_enabled(pmarking(marking), transition)

    @lru_cache
    def _transition_is_enabled(self, marking: PMarking, transition: Transition) -> bool:
        return Occurrence(self, marking, transition).is_enabled()

    def transition_is_external(self, transition: Transition) -> bool:
        """Returns True if a given `Transition` is external in relation to this net."""
        return not self.node_outputs.get(transition) or not self.node_inputs.get(transition)

    def _reset_caches(self):
        self._colors.cache_clear()
        self._marking_after_transition.cache_clear()
        self._transition_is_enabled.cache_clear()
        self.subnet.cache_clear()



from __future__ import annotations

import json
from typing import Mapping

from django.conf import settings
from django.http import Http404
from django.shortcuts import render
from pyrsistent import pmap

from carladam import PetriNet
from carladam.django.petrinet_simulator.marking import decode_marking_from_json


def index(request, petrinets: Mapping[str, PetriNet] | None = None):
    if petrinets is None:
        petrinets = getattr(settings, "CARLADAM_SIMULATOR_PETRINETS", {})
    context = {"petrinets": petrinets}
    return render(request, "petrinet_simulator/index.html", context)


def simulator(request, net_name: str, petrinets: Mapping[str, PetriNet] | None = None):
    if petrinets is None:
        petrinets = getattr(settings, "CARLADAM_SIMULATOR_PETRINETS", {})

    net = petrinets.get(net_name)
    if net is None:
        raise Http404()

    colors = pmap((color.label, color) for color in net.colors)

    # Get options
    rotate = request.GET.get("rotate") == "1"

    # Decode initial marking from JSON.
    initial_marking_str = request.GET.get("initial_marking", "{}")
    initial_marking_json = json.loads(initial_marking_str)
    initial_marking = decode_marking_from_json(net=net, colors=colors, marking_json=initial_marking_json)

    # Decode transition list.
    transition_ids_str = request.GET.get("transitions", "")
    transition_ids = list(transition_ids_str.split(","))
    transitions = []
    transitions_by_id = {t.id: t for t in net.transitions}
    for transition_id in transition_ids:
        if not transition_id:
            continue
        transition = transitions_by_id[transition_id]
        transitions.append(transition)

    # Find current marking after transitions.
    current_marking = initial_marking
    for transition in transitions:
        current_marking = net.marking_after_transition(current_marking, transition)

    enabled_transitions = list(sorted(net.enabled_transitions(current_marking)))
    enabled_subnet = PetriNet.new(*(net.subnet(transition) for transition in enabled_transitions))

    context = dict(
        current_marking=current_marking,
        current_marking_items=list(sorted(current_marking.items())),
        enabled_subnet=enabled_subnet,
        enabled_transitions=enabled_transitions,
        example_markings=getattr(net.Structure, "example_markings", {}),
        initial_marking=initial_marking,
        initial_marking_items=list(sorted(initial_marking.items())),
        net=net,
        net_name=net_name,
        transition_ids=",".join(transition.id for transition in transitions),
        transitions=transitions,
        rotate=rotate,
    )
    return render(request, "petrinet_simulator/simulator.html", context)



import contextlib
import importlib.util
import sys
from pathlib import Path
from typing import Any, Dict, Type
from uuid import uuid4

from decouple import Csv, config
from django.conf import settings
from django.core import management
from django.http import HttpResponse
from django.urls import include, path
from django.utils import module_loading

from carladam import PetriNet


def configure(**kwargs):
    params = dict(
        ALLOWED_HOSTS=config("ALLOWED_HOSTS", default="*", cast=Csv()),
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": ":memory:",
            },
        },
        DEBUG=config("DEBUG", default=True),
        DEFAULT_AUTO_FIELD="django.db.models.AutoField",
        INSTALLED_APPS=["carladam.django.petrinet_simulator"],
        ROOT_URLCONF=__name__,
        SECRET_KEY=config("SECRET_KEY", default="SECRET"),
        TEMPLATES=[
            {
                "BACKEND": "django.template.backends.django.DjangoTemplates",
                "DIRS": [],
                "APP_DIRS": True,
                "OPTIONS": {
                    "context_processors": [
                        "django.template.context_processors.debug",
                        "django.template.context_processors.request",
                    ],
                },
            },
        ],
    )
    return settings.configure(**params, **kwargs)


def home(request):
    return HttpResponse("Welcome!")


urlpatterns = [
    path("", include("carladam.django.petrinet_simulator.urls")),
]


def is_petrinet_subclass(obj: Any) -> bool:
    return isinstance(obj, type) and issubclass(obj, PetriNet)


def main():
    configure()

    petrinets: Dict[str, PetriNet] = {}
    settings.CARLADAM_SIMULATOR_PETRINETS = petrinets

    arg0, *args = sys.argv

    if "--" in args:
        separator_index = args.index("--")
        runserver_args, simulator_args = args[:separator_index], args[separator_index + 1 :]
    else:
        runserver_args, simulator_args = [], args

    for arg in simulator_args:
        with contextlib.suppress(ImportError, ValueError):
            petrinet_cls = module_loading.import_string(arg)
            if is_petrinet_subclass(petrinet_cls):
                petrinets[petrinet_cls.__name__] = petrinet_cls.new()
                continue
        arg_path = Path(arg)
        if not arg_path.exists():
            print(f"Warning: Could not find {arg_path!r}")
            continue
        if arg_path.is_dir():
            paths = arg_path.glob("**/*.py")
        elif arg_path.suffix != ".py":
            print(f"Warning: Not a Python module {arg_path!r}")
            continue
        else:
            paths = [arg_path]
        for module_path in sorted(paths):
            with contextlib.suppress(ImportError):
                # cf. https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
                module_name = f"__{uuid4().hex}__"
                spec = importlib.util.spec_from_file_location(module_name, module_path)
                module = importlib.util.module_from_spec(spec)
                sys.modules[module_name] = module
                spec.loader.exec_module(module)
                for key in sorted(dir(module)):
                    petrinet_cls: Type[PetriNet] = getattr(module, key)
                    if petrinet_cls is PetriNet:
                        continue
                    if not is_petrinet_subclass(petrinet_cls):
                        continue
                    petrinets[petrinet_cls.__name__] = petrinet_cls.new()

    print("")
    print("PetriNet classes loaded:")
    print("\n".join(f"- {key}" for key in sorted(petrinets)))
    print("")

    management.execute_from_command_line([arg0, "runserver", *runserver_args])


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import AbstractSet, TYPE_CHECKING, overload

from attr import define, field
from pyrsistent import PMap, PSet

from carladam.petrinet import defaults
from carladam.petrinet.color import Color, ColorSet
from carladam.petrinet.defaults import default_id

if TYPE_CHECKING:  # pragma: nocover
    from carladam.petrinet.arc import ArcPT, ArcTP, CompletedArcPT
    from carladam.petrinet.transition import Transition


@define
class Place:
    """A `Place` represents where `Token`s can be mapped to via a `Marking` of a `PetriNet`."""

    id: str = field(factory=default_id, repr=False)
    "Globally unique ID, used as a hash key."

    name: str = field()
    "Descriptive name, usually but not necessarily unique within a net."

    icon: str | None = defaults.PLACE
    "Icon/emoji to use when decorating this transition visually."

    # noinspection PyUnresolvedReferences
    @name.default
    def _default_name_is_id(self):
        """By default, `name` is set to the `id` if not passed in."""
        return self.id

    def __hash__(self):
        return hash(self.id)

    def __lt__(self, other):
        """Places are sorted by name."""
        return self.name < other.name

    def __repr__(self):
        if self.icon:
            return f"{self.icon} {self.name}"
        if self.name != self.id:
            return self.name
        return f"<Place id={self.id!r}>"

    def __lshift__(self, other: Transition | Color | ColorSet | AbstractSet[Color]) -> ArcTP:
        """Returns an arc from this `Place` from the given `Transition`."""

        from carladam.petrinet.arc import ArcTP, CompletedArcTP
        from carladam.petrinet.transition import Transition

        if isinstance(other, Color):
            other = {other: 1}
        if isinstance(other, (set, frozenset, PSet)):
            other = {color: 1 for color in other}
        if isinstance(other, (dict, PMap)):
            return ArcTP(src=None, dest=self, weight=other)

        if isinstance(other, Transition):
            return CompletedArcTP(src=other, dest=self)

        raise TypeError("Place cannot be connected to object.", other)

    @overload
    def __rshift__(self, other: Transition) -> CompletedArcPT: ...

    @overload
    def __rshift__(self, other: Color | ColorSet | AbstractSet[Color]) -> ArcPT: ...

    def __rshift__(self, other: Transition | Color | ColorSet | AbstractSet[Color]) -> ArcPT | CompletedArcPT:
        """Returns an arc from this `Place` to the given `Transition`."""

        from carladam.petrinet.arc import ArcPT, CompletedArcPT
        from carladam.petrinet.transition import Transition

        if isinstance(other, Color):
            other = {other: 1}
        if isinstance(other, (set, frozenset, PSet)):
            other = {color: 1 for color in other}
        if isinstance(other, (dict, PMap)):
            return ArcPT(src=self, dest=None, weight=other)

        if isinstance(other, Transition):
            return CompletedArcPT(src=self, dest=other)

        raise TypeError("Place cannot be connected to object.", other)



from __future__ import annotations

from typing import AbstractSet, Any, Callable, Mapping

from attr import Factory, define, field
from pyrsistent import pmap, pset

from carladam.petrinet.color import Abstract, Color
from carladam.petrinet.defaults import default_id


@define
class Token:
    """A `Token` represents a real or abstract object used to model state of `Place`s and I/O of `Transitions`."""

    id: str = field(factory=default_id, repr=False)
    "Globally unique ID, used as a hash key."

    name: str = field()
    "Descriptive name, usually but not necessarily unique within a net."

    color: Color = Abstract
    "Discrete label describing the kind of this token."

    data: Mapping[str, Any] = Factory(pmap)
    "Arbitrary data attached to the token, potentially used during the `guard` or `fn` of a `Transition`."

    # noinspection PyUnresolvedReferences
    @name.default
    def _default_name_is_id(self):
        """By default, `name` is set to the `id` if not passed in."""
        return self.id

    def replace(self, **kwargs) -> Token:
        """Return a copy of this `Token` but with `data` replaced with `kwargs`."""
        return Token(id=self.id, name=self.name, color=self.color, data=kwargs)

    def clone(self) -> Token:
        """Return a clone of this `Token` but having a new `id` and `name`."""
        return Token(id=default_id(), color=self.color, data=self.data)

    def __hash__(self):
        return hash(self.id)

    def __lt__(self, other):
        return (self.color.label, self.name) < (other.color.label, other.name)

    def __mul__(self, quantity: int) -> TokenSet:
        if not isinstance(quantity, int):
            raise TypeError("Can only multiply tokens by integers")
        return frozenset(self.clone() for _ in range(quantity))

    def __repr__(self):
        if self.name == self.id:
            s = repr(self.color)
        else:
            s = f"{self.color!r} {self.name}"
        if self.data:
            value_list = [f"{key}={value!r}" for key, value in self.data.items()]
            return f"{s}({', '.join(value_list)})"
        return s


TokenSet = AbstractSet[Token]
"A set of `Token`."

TokenFilter = Callable[[TokenSet], TokenSet]
TokenReducer = Callable[[TokenSet], Token]


def tokens_where(*conditions) -> TokenFilter:
    """
    Return a function that filters a set of tokens to only those that match all conditions given.

    Each condition is a function that takes a token and returns True or False.
    """

    def where_token_filter(tokens: TokenSet) -> TokenSet:
        return pset(token for token in tokens if all(condition(token) for condition in conditions))

    return where_token_filter


def _no_filter(tokens: TokenSet) -> TokenSet:
    return tokens


def one(token_filter: TokenFilter = _no_filter) -> TokenReducer:
    """Return a function taking a set of one token, returning the token or raising `ValueError` if >1 token given."""

    def one_token_reducer(tokens: TokenSet) -> Token:
        tokens = token_filter(tokens)
        if len(tokens) != 1:
            raise ValueError(f"Expected 1 token, got {len(tokens)}")
        return next(iter(tokens))

    return one_token_reducer
